//	spearsynth1~ -- use sine waves to do a crude SPEAR resynth
//			Brad Garton, 12/2005


#include "ext.h"
#include "z_dsp.h"
#include "string.h"
#include "ext_strings.h"
#include "edit.h"
#include "ext_wind.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#define MAX_OUTPUTS 2	// also arbitrary
// we are total memory hogs with the #defines below -- this could be handled much much better
#define MAX_PARTIALS 500 // this is how many total partials we can read in at any one time.
#define MAX_FRAMES 1000 // allow for 10 seconds

// oh I am a bad brad programmer here!  These *need* to be allocatedm and inside the t_spearsynth1 structure
// right now only 1 spearsynth1~ will operate
float si[MAX_FRAMES][MAX_PARTIALS]; // store the sampling increments used for the oscillator frequency
float phase[MAX_FRAMES][MAX_PARTIALS]; // store where we currently are located in the sinetable[]
float amp[MAX_FRAMES][MAX_PARTIALS]; // amplitudes for the partials

void *spearsynth1_class;

// declare the sine wave table out here so that all instances of the spearsynth1~ object
// can reference it; that way we don't create it anew for each object.
#define MAXSINETABLE 44100 // make it easy for us -- 1 cycle/second; si then equals freq
float sinetable[MAXSINETABLE];

// in this structure are variables specific to each instance of spearsynth1~
typedef struct _spearsynth1
{
	// header
	t_pxobject x_obj;

	long num_outputs; 	// number of outputs (mono or stereo)

	float curtime; // the current time (in seconds) through the synthesis
	float nexttime; // when to shift to the next frame of SPEAR data (usually curtime + 0.01 seconds)
	int goflag; // when "1", synthesis starts.  Set by a bang.
	int numframes; // total number of frames to be used
	int curframe; // current frame data being used
	int data_read; // set to "1" when data has been read
} t_spearsynth1;


/****PROTOTYPES****/

//	max/msp setup funcs; this probably won't change, unless you decide to change the number of
//	args that the user can input, in which case spearsynth1_new will have to change
void *spearsynth1_new(long num_outputs);	// this allows us to have one optional params for the object
void spearsynth1_dsp(t_spearsynth1 *x, t_signal **sp, short *count); 
t_int *spearsynth1_perform(t_int *w);
void spearsynth1_assist(t_spearsynth1 *x, void *b, long m, long a, char *s);

// the following are functions we define below for handling floats/ints/bangs
// at the inputs.  We can delete these if we don't use them
void spearsynth1_float(t_spearsynth1 *x, double f);
void spearsynth1_int(t_spearsynth1 *x, int i);
void spearsynth1_bang(t_spearsynth1 *x);

// the functions for reading in the SPEAR data file -- doread() is so we can defer()
void spearsynth1_read(t_spearsynth1 *x, Symbol *s, short argc, Atom *argv);
void spearsynth1_doread(t_spearsynth1 *x, Symbol *s, short argc, t_atom *argv);

/****FUNCTIONS****/

//primary MSP funcs -- this is called the first time you instantiate an object of this type
void main(void)
{
	int i;
	float radians;
	
	// the A_DEFLONG argument sets mono or stereo output (default will be mono)
	setup((struct messlist **)&spearsynth1_class, (method)spearsynth1_new, (method)dsp_free, (short)sizeof(t_spearsynth1), 0L, A_DEFLONG, 0);

	// max/msp standard messages; don't change these  
	addmess((method)spearsynth1_dsp, "dsp", A_CANT, 0);
	addmess((method)spearsynth1_assist,"assist", A_CANT,0);

	// spearsynth1_float will be called for any floats coming in any inlet
	addfloat((method)spearsynth1_float);

	// spearsynth1_int will be called for any ints coming in any inlet
	addint((method)spearsynth1_int);

	// spearsynth1_bang will be called for any bangs coming in any inlet
	addbang((method)spearsynth1_bang);

	// this is for the [read ...] message
	addmess((method)spearsynth1_read, "read", A_GIMME, 0);

	// msp initialization (gotta have this)
	dsp_initclass();

	// set up the sine wavetable
	for (i = 0; i < MAXSINETABLE; i++) {
		radians = (float)i/(float)MAXSINETABLE * TWOPI;
		sinetable[i] = sin(radians);
	}

	post("spearsynth1~ external loaded!");
}


// this gets called when the object is created and every time the user types in new args
void *spearsynth1_new(long num_outputs)
{
	int i,j;

	// leave this; creates our object
	t_spearsynth1 *x = (t_spearsynth1 *)newobject(spearsynth1_class);

	// zero out the struct, to be careful (takk to jkclayton)
	if (x) { 
		for (i = sizeof(t_pxobject); i < sizeof(t_spearsynth1); i++)  
			((char *)x)[i] = 0;
	}

	// this is where we access the values (if any) of the optional params for the object	
	// constrain number of outputs
	if (num_outputs < 1) num_outputs = 1;
	if (num_outputs > MAX_OUTPUTS) num_outputs = 1;

	// and we set the internal t_spearsynth1 structure variables
	x->num_outputs = num_outputs;
	x->goflag = 0; // "1" starts synthesis output, from a bang
	x->data_read = 0; // set to "1" after reading data file

	// now setup audio inputs and outputs
	// inputs
	dsp_setup((t_pxobject *)x, 0);
   
	// signal outputs
	for (i = 0; i < x->num_outputs; i++) {
		outlet_new((t_object *)x, "signal");
	}

	//occasionally this line is necessary if you are doing weird asynchronous things with the in/out vectors
	//x->x_obj.z_misc = Z_NO_INPLACE;

	return (x);
}


// this gets called everytime audio is started; even when audio is running, if the user changes anything
// (like deletes a patch cord), audio will be turned off and then on again, calling this func.
// this adds the "perform" method to the DSP chain, and also tells us
// where the audio vectors are and how big they are
void spearsynth1_dsp(t_spearsynth1 *x, t_signal **sp, short *count)
{
	void *dsp_add_args[MAX_OUTPUTS + 2];
	int i;

	// construct the array of vectors and stuff for max/msp to knoww
	dsp_add_args[0] = x; // the object itself
	for (i = 0; i < (x->num_outputs); i++) { // pointers to the input and output vectors
		dsp_add_args[i+1] = sp[i]->s_vec;
	}
	dsp_add_args[x->num_outputs + 1] = (void *)sp[0]->s_n; // pointer to the vector size
	dsp_addv(spearsynth1_perform, (x->num_outputs + 2), dsp_add_args); // add them to the signal chain

	x->goflag = 0; // "1" starts synthesis output, from a bang
}


// this is where the action is
// we get vectors of samples (n samples per vector), process them and send them out
// the "w" array contains a lot of stuff (note no input pointers in this one!)
//	w[0] -- pointer to the spearsynth1_perform function being called
//	w[1] -- pointer to the t_spearsynth1 object structure data
//	w[2] ... w[numoutputs+1] -- pointers to arrays to be filled with samples (output arrays)
//	w[numoutputs+2] -- number of samples/vector (i.e. the number of samples in each input or output array)
t_int *spearsynth1_perform(t_int *w)
{
	float timeadd;
	t_spearsynth1 *x = (t_spearsynth1 *)(w[1]);

	float *out[MAX_OUTPUTS];	// pointers to the output vectors
	// note the the above are pointers -- each output is represented as an array
	// of samples (i.e. non-interleaved).  These are set to point to the w[] arrays.

	long n = w[x->num_outputs + 2];	// number of samples per vector

	// local vars
	int i;

	// check to see if we should skip this routine if the patcher is "muted"
	// all audio objects should have this, even though the 'goto' is supremely elegant.
	if (x->x_obj.z_disabled) goto out;

	// assign the output vector pointers
	for (i = 0 ;i < x->num_outputs; i++) {
		out[i] = (float *)(w[i+2]);
	}
	// after this loop above, we can write samples using *out[oulet]+sampval to
	// get and set sample values for each outlet vector.


/* ------- this is the sample computing loop for each vector ------- */
	timeadd = 1.0/44100.0; // amount to add for each sample processed
	while (n--) { // n is the number of samples for the vector

		*out[0] = 0.0;
		if (x->goflag == 1) {
			for (i = 0; i < MAX_PARTIALS; i++) {
				if (amp[x->curframe][i] > 0.0) {
/*
					*out[0] += (sinetable[(int)phase[x->curframe][i]] * amp[x->curframe][i]);
					phase[x->curframe][i] += si[x->curframe][i];
					if (phase[x->curframe][i] > (float)MAXSINETABLE)
						phase[x->curframe][i] = phase[x->curframe][i] - (float)MAXSINETABLE;
*/
					*out[0] += (sinetable[(int)phase[0][i]] * amp[x->curframe][i]);
					phase[0][i] += si[x->curframe][i];
					if (phase[0][i] > (float)MAXSINETABLE)
						phase[0][i] = phase[0][i] - (float)MAXSINETABLE;
				}
			}

			x->curtime += timeadd;
			if (x->curtime >= x->nexttime) {
				x->curframe++;
				x->nexttime = x->curtime + 0.1;
				if (x->curframe >= MAX_FRAMES) {
					x->goflag = 0;
				}
			}
		}
		
		if (x->num_outputs == 2) *out[1] = *out[0];
		*out[0]++;
		*out[1]++;
	}
/* ------- end of the sample computing loop ------- */


	// return a pointer to the next object in the signal chain.
out:
	return w + x->num_outputs + 3;
}


// tells the user about the inputs/outputs when mousing over them
// "m" is inlet (1) or outlet (2); "a" is which one the mouse is over
void spearsynth1_assist(t_spearsynth1 *x, void *b, long m, long a, char *s)
{
	int i;
	
	//could use switch/case inside for loops, to give more informative assist info....
	if (m == 1) sprintf(s, "bang to trigger");  // this is the message that gets shown in the window frame
	if (m == 2) {
		for (i = 0; i <= x->num_outputs; i++) // "<=" because we have one additional outlet for bangs/vals
			if (a == i) sprintf(s, "signal %d", i);  // ditto about message that gets shown
	}
}


// this gets called whenever a float is received at *any* input
void spearsynth1_float(t_spearsynth1 *x, double f)
{
	// not using float inputs in this version
}


// this gets called whenever an int is received at *any* input
void spearsynth1_int(t_spearsynth1 *x, int k)
{
	// not using int inputs in this version
}


// we're going to use the bang to reset to the beginning of the SPEAR data and trigger synthesis
// this gets called whenever a bang is received at *any* input
void spearsynth1_bang(t_spearsynth1 *x)
{
	if (x->data_read == 0) {
		post("spearsynth1~: you must read in a SPEAR data file before running");
		return;
	}

	// start at the beginning
	x->curtime = 0.0;
	x->nexttime = 0.01; // for SPEAR resampleframes data
	x->curframe = 0;
	x->goflag = 1;  // and go!
}


// the [read] message reads in a SPEAR resampleframes text file, and sets a pointer to the
// memory for use in the spearsynth1_perform() function.  The read gets deferred because it
// is a slow operation
void spearsynth1_read(t_spearsynth1 *x, Symbol *s, short argc, Atom *argv)  
{
	defer(x, (method)spearsynth1_doread, s, argc, argv); // always defer this message  
}

// this opens and reads in a file
void spearsynth1_doread(t_spearsynth1 *x, Symbol *s, short argc, t_atom *argv) 
{
	char filename[256];
	int nameflag; // we use this to determine if we should open a dialog
	short err, i, j;
	short path;
	long type = 'TEXT'; 
	long size;
	long outtype, count;
	t_filehandle fh; // this to reference the SPEAR data file
	t_handle speardatahandle; // use this to reference the SPEAR data we read in

	// these are for the data-reading/parsing ugliness
	char *bufptr;
	char parsebuf[1000];
	int linepartials, pno;
	
	nameflag = 0;	
	for (i = 0; i < argc; i++) {
		switch (argv[i].a_type) {
			case A_LONG:
				post("spearsynth1 read should be a single message or contain a filename");
				return;
			case A_FLOAT:
				post("spearsynth1 read should be a single message or contain a filename");
				return;
			case A_SYM:
				strcpy(filename, argv[i].a_w.w_sym->s_name);
				nameflag = 1;
		}
	}
	
	if (nameflag == 0) {
		if (open_dialog(filename,  &path, &outtype, 0L, 0)) // allow all types of files
			return; //user cancelled
	} else {
		if (locatefile_extended(filename, &path, &outtype, &type, 1)) {
			error("spearsynth1~: error opening file: can't find file"); 
			return; //not found
		}
	}
	
	//we should have a valid filename at this point
	err = path_opensysfile(filename, path, &fh, READ_PERM);  
	if (err) {       
		error("error %d opening file", err); 
		return;
	}
	
	sysfile_geteof(fh, &size); // get the size of the file
	if (!(speardatahandle = sysmem_newhandle(size))) { // allocate the memory
		error("spearsynth1~: %s too big to read", filename);
		return;
	} else {
		err = sysfile_readtextfile(fh, speardatahandle, size, TEXT_LB_NATIVE); // read that file!
		if (err) {
			error("trouble %d reading SPEAR data file", err);
			return;
		}
	}

	// now we're going to go through the data we just read and grab the info we need
	// and store them in our arrays.  jeez, this is truly insane...
	bufptr = (char *)*speardatahandle; // bufptr initially set to the beginning of the data
	
	for (i = 0; i < 8; i++) {
		sscanf(bufptr, "%s ", parsebuf);
		bufptr += (strlen(parsebuf)+1); // get the 'white space'
	}

	// the next item should be the frame-count number
	sscanf(bufptr, "%s", parsebuf);
	bufptr += (strlen(parsebuf)+1);
	sscanf(parsebuf, "%d", &(x->numframes));
	if (x->numframes > MAX_FRAMES) {
		post("spearsynth1~ can only do %d frames of data, limiting to that amount", MAX_FRAMES);
		x->numframes = MAX_FRAMES;
	}

	// go one more...
	sscanf(bufptr, "%s", parsebuf);
	bufptr += (strlen(parsebuf)+1);

	// now we should be positioned at the beginning of the frame data

	// initialize amp and phase for this data -- we use amp to determine synthesis
	for (i = 0; i < MAX_FRAMES; i++) {
		for (j = 0; j < MAX_PARTIALS; j++) {
			amp[i][j] = 0.0;
		}
	}

	// now get the amp and freq (si) values for the appropriate partials
	for (i = 0; i < x->numframes; i++) {
		// read the frame time-val (unused) and the number of partials for this line
		sscanf(bufptr, "%s", parsebuf);
		bufptr += (strlen(parsebuf)+1);
		sscanf(bufptr, "%s", parsebuf);
		bufptr += (strlen(parsebuf)+1);
		sscanf(parsebuf, "%d", &linepartials);

		for (j = 0; j < linepartials; j++) {
			// get the partial #
			sscanf(bufptr, "%s", parsebuf);
			bufptr += (strlen(parsebuf)+1);
			sscanf(parsebuf, "%d", &pno);
			// and the freq (si)
			sscanf(bufptr, "%s", parsebuf);
			bufptr += (strlen(parsebuf)+1);
			if (pno < MAX_PARTIALS) {
				sscanf(parsebuf, "%f", si[i]+pno);
			} else {
				post("spearsynth1~ can only do %d partials, limiting to that amount", MAX_PARTIALS);
			}
			// and the amp
			sscanf(bufptr, "%s", parsebuf);
			bufptr += (strlen(parsebuf)+1);
			if (pno < MAX_PARTIALS) {
				sscanf(parsebuf, "%f", amp[i]+pno);
			}
			// and set the phase to 0 for this one
			if (pno < MAX_PARTIALS) {
//				phase[i][pno] = 0.0;
				phase[0][pno] = 0.0;

			}
		}
	}

	// free up the resources we used for this read operation
	sysmem_freehandle(speardatahandle);
	sysfile_close(fh);

	x->data_read = 1; // all set!

	return;
}
